package knightminer.inspirations.library.recipe.cauldron.contenttype;

import com.google.gson.JsonObject;
import com.google.gson.JsonSyntaxException;
import io.netty.handler.codec.DecoderException;
import io.netty.handler.codec.EncoderException;
import knightminer.inspirations.library.recipe.cauldron.CauldronContentTypes;
import knightminer.inspirations.library.recipe.cauldron.contents.ICauldronContents;
import net.minecraft.nbt.CompoundNBT;
import net.minecraft.network.PacketBuffer;
import net.minecraft.util.JSONUtils;
import net.minecraftforge.common.util.Constants.NBT;

import javax.annotation.Nullable;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;

/**
 * Represents a type of contents that can be stored in the cauldron
 * @param <C>  {@link ICauldronContents} implementation for this type
 */
public abstract class CauldronContentType<C extends ICauldronContents, T> {
  private final Map<T, C> cache = new HashMap<>();
  private final Class<C> clazz;
  private final Function<? super T, ? extends C> constructor;
  private final Function<C, T> valueGetter;
  private final String key;

  /**
   * Creates a new instance
   * @param clazz  Content type class for validation
   * @param constructor  Constructor to create contents from the type
   * @param valueGetter  Function to get the value from a content type
   * @param key          Key to use for serializing and deserializing
   */
  protected CauldronContentType(Class<C> clazz, Function<? super T,? extends C> constructor, Function<C,T> valueGetter, String key) {
    this.clazz = clazz;
    this.constructor = constructor;
    this.valueGetter = valueGetter;
    this.key = key;
  }

  /**
   * Creates a new instance using "name" as the serialization key
   * @param clazz  Content type class for validation
   * @param constructor  Constructor to create contents from the type
   * @param valueGetter  Function to get the value from a content type
   */
  protected CauldronContentType(Class<C> clazz, Function<? super T,? extends C> constructor, Function<C,T> valueGetter) {
    this(clazz, constructor, valueGetter, "name");
  }

  /* Base methods */

  /**
   * Checks if the given contents matches this type
   * @param contents  Contents to check
   * @return True if the type matches
   */
  public boolean matches(ICauldronContents contents) {
    return clazz.isInstance(contents);
  }

  /**
   * Gets the contents as this type
   * @param contents  Contents to fetch
   * @return  Type to get
   */
  public Optional<T> get(ICauldronContents contents) {
    if (clazz.isInstance(contents)) {
      return Optional.of(clazz.cast(contents)).map(valueGetter);
    }
    return Optional.empty();
  }


  /* Creation methods */

  /**
   * Adds an override to this type, preventing the default constructor
   * @param value     Override to fetch
   * @param instance  Instance to use for override
   */
  public void addOverride(T value, C instance) {
    cache.put(value, instance);
  }

  /**
   * Gets a value of the given type
   * @param value  Type of value
   * @return  Value to fetch
   */
  public C of(T value) {
    return cache.computeIfAbsent(value, constructor);
  }


  /* Serializing and deserializing */

  /**
   * Gets the name of the given value for writing to JSON and NBT
   * @param value  Value
   * @return  Name of the value
   */
  public abstract String getName(T value);

  /**
   * Gets the entry for a given value
   * @param name  Name
   * @return  Entry, or null if missing
   */
  @Nullable
  public abstract T getEntry(String name);

  /**
   * Gets the key used for JSON and NBT
   * @return  JSON and NBT key
   */
  public String getKey() {
    return key;
  }

  /**
   * Reads the given type from NBT
   * @param tag  NBT tag
   * @return  Read value
   */
  @Nullable
  public C read(CompoundNBT tag) {
    if (tag.contains(key, NBT.TAG_STRING)) {
      T value = getEntry(tag.getString(key));
      if (value != null) {
        return of(value);
      }
    }
    return null;
  }

  /**
   * Reads the given type from JSON
   * @param json  JSON object
   * @return  Read value=
   * @throws com.google.gson.JsonSyntaxException if the JSON is invalid
   */
  public C read(JsonObject json) {
    String name = JSONUtils.getString(json, key);
    T value = getEntry(name);
    if (value == null) {
      throw new JsonSyntaxException("Invalid name '" + name + "' for type '" + CauldronContentTypes.getName(this) + "'");
    }
    return of(value);
  }

  /**
   * Reads the given type from the packet buffer
   * @param buffer  Packet buffer
   * @return  Read value
   * @throws DecoderException if the type is invalid
   */
  public C read(PacketBuffer buffer) {
    String name = buffer.readString(Short.MAX_VALUE);
    T value = getEntry(name);
    if (value != null) {
      return of(value);
    }
    throw new DecoderException("Invalid name '" + name + "' for type " + this);
  }

  /**
   * Gets the name of the given value
   * @param value  Value
   * @return  Name of the value
   */
  private String getName(C value) {
    return getName(valueGetter.apply(value));
  }

  /**
   * Writes the given type to NBT
   * @param contents  Contents to write
   * @param tag       NBT tag
   */
  public void write(ICauldronContents contents, CompoundNBT tag) {
    if (clazz.isInstance(contents)) {
      tag.putString(key, getName(clazz.cast(contents)));
    }
  }

  /**
   * Writes the given type to JSON
   * @param contents  Contents to write
   * @param json      JSON object
   */
  public void write(ICauldronContents contents, JsonObject json) {
    if (clazz.isInstance(contents)) {
      json.addProperty(key, getName(clazz.cast(contents)));
    }
  }

  /**
   * Writes the given type to the packet buffer
   * @param contents  Contents to write
   * @param buffer    Packet buffer
   */
  public void write(ICauldronContents contents, PacketBuffer buffer) {
    if (clazz.isInstance(contents)) {
      buffer.writeString(getName(clazz.cast(contents)));
    } else {
      throw new EncoderException("Invalid type for contents argument");
    }
  }

  @Override
  public String toString() {
    return String.format("CauldronContentType[%s]", CauldronContentTypes.getName(this));
  }
}
